#include "z3-api/c++/z3++.h"
#include <string>
#include <map>
#include <iostream>
#include <vector>

#define STR(x, y) \
((std::string(x) + (y)).c_str())

typedef std::map<std::string, z3::expr> state_store;

// Normalizes Booleans and 1-bit bitvectors into Booleans.
z3::expr z3_bool(const z3::expr& x)
{
    if (x.is_bool()) return x;
    if (x.is_bv() and x.get_sort().bv_size() == 1) return x == x.ctx().bv_val(1, 1);
    assert(false);
}

void create_state(z3::context& ctx, int num, state_store& state)
{
    std::string suffix = "@" + std::to_string(num);
    // Create entries in state that map from name to a variable.
    // The name of the variable must be unique, so use STR(name, suffix) for this.
    state.emplace("reset", ctx.bv_const(STR("reset", suffix), 1));
    state.emplace("din", ctx.bv_const(STR("din", suffix), 8));
    state.emplace("dout", ctx.bv_const(STR("dout", suffix), 8));
    state.emplace("init", ctx.bv_const(STR("init", suffix), 1));
    state.emplace("uut.buffer", ctx.bv_const(STR("uut.buffer", suffix), 8));
    state.emplace("uut.state", ctx.bv_const(STR("uut.state", suffix), 2));
}

void init_state(z3::context& ctx, state_store& state, z3::expr_vector& conds)
{
    // Create constraints related to the initial state of the circuit.
    conds.push_back(state.at("init") == 1);
}

void get_assumes(z3::context& ctx, state_store& state, z3::expr_vector& conds)
{
    // Create the assumptions that control the environment.
    z3::expr init_bool = z3_bool(state.at("init"));
    z3::expr reset_bool = z3_bool(state.at("reset"));

    conds.push_back(z3::implies(init_bool, reset_bool));
}

void get_asserts(z3::context& ctx, state_store& state, z3::expr_vector& conds)
{
    // Create the assertions that check safety properties.
    z3::expr reset_bool = z3_bool(state.at("reset"));
    z3::expr dout_bits = state.at("dout").extract(1, 0);
    conds.push_back(z3::implies(!reset_bool, dout_bits == 0));
}

void get_trans(z3::context& ctx, state_store& state0, state_store& state1, z3::expr_vector& conds)
{
    // Create the transition constraint over the two states.
    conds.push_back(state1.at("init") == 0);

    z3::expr state_reset = ctx.bv_val(0, 2);
    z3::expr state_0 = ctx.bv_val(1, 2);
    z3::expr state_1_then = state0.at("uut.state");
    z3::expr state_1_else = ctx.bv_val(2, 2);
    z3::expr cond = state0.at("uut.buffer").extract(1, 0) != 0;
    z3::expr state_1 = z3::ite(cond, state_1_then, state_1_else);
    z3::expr state_2 = ctx.bv_val(0, 2);

    z3::expr old_state = state0.at("uut.state");

    z3::expr reset_bool = z3_bool(state0.at("reset"));

    z3::expr ns_state = z3::ite(reset_bool,
                                state_reset,
                                z3::ite(old_state == 0,
                                        state_0,
                                        z3::ite(old_state == 1,
                                                state_1, state_2))
    );
    conds.push_back(state1.at("uut.state") == ns_state);

    z3::expr dout_reset = ctx.bv_val(0, 8);
    z3::expr dout2 = state0.at("dout") + state0.at("uut.buffer");
    z3::expr dout01 = state0.at("dout");

    z3::expr ns_dout = z3::ite(reset_bool,
                               dout_reset,
                               z3::ite(old_state != 2, dout01, dout2));

    conds.push_back(state1.at("dout") == ns_dout);

    z3::expr buffer0 = state0.at("din");
    z3::expr buffer1_then = state0.at("uut.buffer") + 1;
    z3::expr buffer1_else = state0.at("uut.buffer");
    z3::expr buffer1 = z3::ite(cond, buffer1_then, buffer1_else);
    z3::expr buffer2 = state0.at("uut.buffer");

    z3::expr ns_buffer =  z3::ite(reset_bool, state0.at("uut.buffer"),
                                  z3::ite(old_state == 0, buffer0,
                                          z3::ite(old_state == 1, buffer1, buffer2)));

    conds.push_back(state1.at("uut.buffer") == ns_buffer);
}

int check_sat(z3::solver& solver, int cycle, std::vector<state_store>& states)
{
    auto res = solver.check();
    if (res == z3::sat)
    {
        std::cout << "sat @" << cycle << std::endl;
        z3::model model = solver.get_model();
        for (auto& state: states)
        {
            for(auto& tup: state)
            {
                std::cout << tup.second << " = " << model.eval(tup.second, true) << std::endl;
            }
        }
        return 10;
    }

    if (res == z3::unsat)
    {
        std::cout << "unsat @" << cycle << std::endl;
        return 20;
    }

    assert(false);
}

int main() {
    // Create a context and solver.
    z3::context ctx;
    z3::solver solver(ctx);

    std::vector<state_store> states;

    // Create the initial state.
    state_store state_pre;
    create_state(ctx, 0, state_pre);
    states.push_back(state_pre);

    // Assert all the constraints for the initial state.
    z3::expr_vector conds(ctx);
    init_state(ctx, state_pre, conds);
    solver.add(conds);

    // Assert all the assumptions for all states.
    conds.resize(0);
    get_assumes(ctx, state_pre, conds);
    solver.add(conds);

    // Create a new solver frame.
    solver.push();

    // Assert the negation of all assertions.
    conds.resize(0);
    // get_asserts(ctx, state_pre, conds);
    // solver.add(!z3::mk_and(conds));
    int res = check_sat(solver, 0, states);

    // Destroy the solver frame.
    solver.pop();

    // if (res == 10) return 0;

    for (int i = 1; i < 3; i++)
    {
        // Create a new state_store.
        // Assert all the transition conditions from get_trans.
        // Assert the assumptions for every state.
        // Check if the assertions can be violated in a separate frame.
        // If the result was SAT, terminate the program.
        // Otherwise, replace state_pre with the new state you created.
    }

    return 0;
}
